"""
Constants.

Constants for feeds and feed entries.
"""

ATTR_ALERT = "alert"
ATTR_ATTRIBUTION = "attribution"
ATTR_CATEGORY = "category"
ATTR_DESCRIPTION = "description"
ATTR_ID = "id"
ATTR_GUID = "guid"
ATTR_MAG = "mag"
ATTR_PLACE = "place"
ATTR_PUB_DATE = "pubDate"
ATTR_STATUS = "status"
ATTR_TIME = "time"
ATTR_TITLE = "title"
ATTR_TYPE = "type"
ATTR_UPDATED = "updated"

FILTER_CATEGORIES = "categories"
FILTER_MINIMUM_MAGNITUDE = "minimum_magnitude"
FILTER_RADIUS = "radius"
